################################################################################
#               Implementation of several canal objects                        #
#                                                                              #
#                       JAZIRI Samy - 04/04/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#                                                                              #
################################################################################

from time import sleep
from functools import reduce

import threading as th
from decimal import Decimal

def str_to_action(x) :
    return (False,Decimal(x[1:])) if x[0] == '#' else (True,x) 

################################################################################
#### Class STRCanal
################################################################################

class STRCanal :
    '''
    '''

    # Constructor
    def __init__(self,init_str = '') :
        self.string = init_str.split(',') if init_str else []
        self.string = [ str_to_action(x) for x in reversed(self.string) ]
        self.count = len(self.string)
        self.running = False
        
    # Methods
    def pop(self) :
        assert self.running and self.count > 0
        self.count -= 1
        #if not self.count : self.stop()
        return self.string.pop()

    def write(self,b,act) :
        assert self.running
        self.string.append((b,act))
        self.count += 1

    def start(self) :
        self.running = True

    def stop(self) :
        self.running = False

    # Overloading operator
    def __add__(self,action) :
        assert self.running
        self.string.append(str_to_action(action))
        self.count += 1
        
################################################################################
#### Class STDINCanal
################################################################################

class STDINCanal :
    '''
    '''

    # Constructor
    def __init__(self,init_str = '') :
        self.string = init_str.split(',') if init_str else []
        self.string = [ str_to_action(x) for x in reversed(self.string) ]
        self.count = len(self.string)
        self.running = False

    # Methods
    def pop(self) :
        assert self.running and self.count > 0
        self.count -= 1
        return self.string.pop()

    def start(self) :
        self.running = True
        sleep(2)
        while True :
            sleep(2)
            x = input("Enter command (delay begin by #, 'exit' to quit ) : ")
            if x == 'exit' : self.stop(); break 
            else :
                for i in reversed(x.split(',')) : self + i
            
    def stop(self) :
        self.running = False

    # Overloading operator
    def __add__(self,action) :
        assert self.running
        if action :
            self.string.append(str_to_action(action))
            self.count += 1

################################################################################
#### Class Checker
################################################################################

class Checker :
    '''
    '''
    # Constructor
    def __init__(self,atd1,atd2,init_str) :
        self.atd1 = atd1; self.atd2 = atd2
        self.init_string = init_str 
        self.string = init_str.split(',') if init_str else []
        self.string = [ str_to_action(x) for x in reversed(self.string) ]
        self.count = int(bool(init_str))
        self.running = False
        self.ex = False
        self.last_count = 0

    def start(self) :
        self.running = True
        self.count = 1
        timer = 0
        while self.string and self.running :
            if (self.atd1.count != self.atd2.count) : timer += 1
            else : timer = 0
            if timer >= 1000 :
                print('Endind run :',self.init_string)
                stri = ','.join([ ('' if b else '#') + str(a)
                                     for b,a in reversed(self.string) ]) 
                print("Position at the run %s" % stri )
                print("Counts : atd1",self.atd1.count,"atd2",self.atd2.count)
                print("Communication canal :",self.atd1.wcanal.string)
                self.stop()
                self.ex = True
            
            if ( (not self.count) and
                 (self.atd1.count == self.atd2.count) and
                 (self.atd1.count + self.atd1.e_count == self.last_count + 1) ) :
                if self.check() :
                    self.count = min(1,len(self.string))
                    self.last_count += 1
                else :
                    print("Check error for atd" +
                          " %s and atd %s" % (self.atd1.atd.name,
                                              self.atd2.atd.name))
                    print("Run %s" % self.init_string)
                    stri = ','.join([ ('' if b else '#') + str(a)
                                     for b,a in reversed(self.string) ]) 
                    print("Position at the run %s" % stri )
                    print("Configurations : ")
                    print(self.atd1.config)
                    print(self.atd2.config)
                    print(self.atd1.config.reachable_states()[0])
                    print(self.atd2.config.reachable_states()[0])
                    self.stop()
                    self.ex = True
            else :
                sleep(0.01)
        self.stop()

    def stop(self) :
        self.running = False
        self.count = 0
        self.last_count = 0
                
    def check(self) :
        #print("Checking %s and %s" % (self.atd1.atd.name,self.atd2.atd.name))
        r2 = self.atd2.config.reachable_states()[0]
        r1 = self.atd1.config.reachable_states()[0]
        ret = reduce(lambda x,y : x and y,[q in r2 for q in r1], True)
        if not ret :
            print([q in r2 for q in r1])
            print(r2)
            print(r1)
        return ret
    
    def pop(self) :
        assert self.running and self.count > 0
        self.count = 0
        return self.string.pop()

    def set_string(self,init_str) :
        assert not self.running
        self.init_string = init_str
        self.string = init_str.split(',') if init_str else []
        self.string = [ str_to_action(x) for x in reversed(self.string) ]
        self.count = 1
        self.nb = 0

################################################################################
#### Class AutoDelayCanal
################################################################################

class AutoDelayCanal :
    '''
    '''

    # Constructor
    def __init__(self,init_str = '') :
        self.string = init_str.split(',') if init_str else []
        self.string = [ str_to_action(x) for x in reversed(self.string) ]
        self.count = len(self.string)
        self.running = False

    # Methods
    def pop(self) :
        assert self.running and self.count > 0
        self.count -= 1
        return self.string.pop()

    def start(self) :
        self.running = True
        
        def make_delay() :
            while self.running :
                sleep(2)
                self + '#0.1'
                
        th.Thread(target = make_delay).start()
        while self.running :
            sleep(0.5)
            x = input("Enter command ('exit' to quit ) : ")
            if x == 'exit' : self.stop() 
            else :
                for i in reversed(x.split(',')) : self + i
            
    def stop(self) :
        self.running = False

    # Overloading operator
    def __add__(self,action) :
        assert self.running
        if action :
            self.string.append(str_to_action(action))
            self.count += 1
