################################################################################
#                      Library for ATD simulation                              #
#                                                                              #
#                       JAZIRI Samy - 04/04/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#                                                                              #
################################################################################

from TILib.timlib import *
from ATDLib.canal import STRCanal

from time import sleep,time
from sys import stdout

import builtins

################################################################################
#### Class ATDRunner
################################################################################

class ATDRunner :
    '''
    
    '''

    # Constructor
    def __init__(self,atd) : 
        self.config = atd.init_config(); self.atd = atd
        self.wcanal = None
        self.bench = open('Test_bench/'+atd.name+'.bench','w')
        self.c_run = 0
        self.count = 0
        self.e_count = 0

    # Methods
    def wait(self,d) :
        ''' Wait the delay d '''
        t = time()
        self.atd(d,self.config)
        self.count += 1
        self.lt_wait.append(time()-t)
        if self.wcanal : self.wcanal.write(False,d)

    def make(self,l) :
        ''' Make a transition with label l of the automaton '''
        if self.wcanal :
            if self.atd.silent != l : self.count += 1
            else : self.e_count += 1
        t = time()
        if self.atd(l,self.config) :
            self.lt_make.append(time() - t)
            if not self.wcanal :
                if self.atd.silent != l : self.count += 1
                else : self.e_count += 1
            if self.wcanal and self.atd.silent != l : self.wcanal.write(True,l)
            return True
        if self.wcanal :
            if self.atd.silent != l : self.count -= 1
            else : self.e_count -= 1
        self.lt_make.append(time() - t)
        return False

    def write_on_canal(self,canal = None) :
        self.wcanal = canal if canal else STRCanal()
        self.wcanal.start()
        return self.wcanal

    def stop_canal(self) :
        if self.wcanal :
            self.wcanal.stop()
            self.wcanal = None
    
    def read_canal(self,canal) :
        ''' Read the canal to recover what action to make '''
        ret = []
        while canal.count :
            nat,act = canal.pop()
            ret += [act]
            if nat :
                if not self.make(act) : self.e_count += 1
            else : self.wait(act)

        return ret

    def listen(self,canal,delay,print_mode = 0,f = None) :
        ''' While canal is open, read and execute what's written on it every 
        delay second  '''
        #print('Beginning to listen on canal.\n Starting configuration :\n' +
        #      str(self.config) )
        self.lt_make = []
        self.lt_wait = []
        while canal.running :
            ret = self.read_canal(canal)
            if ret :
                self.config.apply_dec()
                print_if(print_mode,self.atd.name,ret,self.config,f)
            sleep(delay)
        #print('Stop listenning' )
        self.c_run += 1
        moy_make = sum(self.lt_make) / len(self.lt_make) if self.lt_make else -1
        moy_wait = sum(self.lt_wait) / len(self.lt_wait) if self.lt_wait else -1
        self.bench.write('Run %i: %.9f;%.9f\n' % (self.c_run,moy_make,moy_wait))
        self.bench.flush()
        if self.wcanal : self.wcanal.stop()
        self.config = self.atd.init_config()
        self.count = 0
        self.e_count = 0
        
def print_if(print_mode,name,ret,config,f = None) :
    if f :
        def print(s) : f.write(s+'\n'); f.flush()
    else :
        def print(s) : builtins.print(s); stdout.flush()

    print('_'*80)
    print(name)
    print('_'*80)
        
    if print_mode == 0 :
        print(('Read on canal : {0} \n' +
               'New Configuration : \n {1}').format(ret,config))
    elif print_mode == 1 :
        print(('Read on canal : {0} \n' +
               'Reached States : \n {1}').format(ret,config.reachable_states()))
    elif print_mode == 2 :
        print('Read on canal : {0} \n'.format(ret))
    else :
        print('Read something on canal')

    print('_'*80)
